from skimage import feature
from skimage import measure, draw
from scipy import optimize
from scipy.optimize import LinearConstraint
import numpy as np
import matplotlib.pyplot as plt
from numpy import float32


class ShapeFitter(object):
    ''' The class will provide a shape fitter on a binary mask. Currently is
    only possible to fit a circular shape by using RANSAC or correlation merit
    functions OR to fit an anular shape by using correlation merit only.
    '''

    def __init__(self, binary_mask):
        self._mask = binary_mask.copy()
        self._params = None
        self._shape_fitted = None
        self._method = None
        self._success = None

    def __repr__(self):
        return "mask shape %s, parameters %f, shapefitted %s, method %s, status %s" % (
            self._mask.shape(), self._params, self._shapefitted, self._method, self._success)

    def parameters(self):
        '''Return parameters estimated form '''
        return self._params

    def mask(self):
        '''Return current mask of ShapeFitter object'''
        return self._mask

    # def fit(self, shape2fit='circle', method='', usecanny=True, **keywords):
    #
    #     if shape2fit == 'circle':
    #         self._shape2fit = measure.CircleModel
    #     else:
    #         raise Exception('Other shape but circle not implemented yet')
    #
    #     if method == '':
    #         self._params = self._fit_correlation(shape2fit=shape2fit, **keywords)
    #     elif method == 'RANSAC':
    #         self._params = self._fit_ransac(**keywords)
    #     else:
    #         raise ValueError('Wrong fitting method specified')

    def fit_circle_ransac(self,
                          apply_canny=True,
                          sigma=3,
                          display=False,
                          **keywords):
        '''Perform a circle fitting on the current mask using RANSAC algorithm

        Parameters
        ----------
            apply_canny: bool, default=True
                apply Canny edge detection before performing the fit.
            sigma: float, default=10
                if apply_canny is True, you can decide the Canny kernel size.
            display: bool, default=False
                it shows the result of the fit.
        '''
        self._shape_fitted = 'circle'
        self._method = 'ransac'
        img = np.asarray(self._mask.copy(), dtype=float)
        img[img > 0] = 128

        edge = img.copy()
        if apply_canny:
            edge = feature.canny(img, sigma)

        coords = np.column_stack(np.nonzero(edge))

        model, inliers = measure.ransac(
            coords, measure.CircleModel,
            keywords.pop('min_samples', 10), residual_threshold=0.01,
            max_trials=1000)
        cx, cy, r = model.params

        if display is True:
            print(r"Cx-Cy {:.2f}-{:.2f}, R {:.2f}".format(cx, cy, r))
            rr, cc = draw.disk((model.params[0], model.params[1]),
                               model.params[2],
                               shape=img.shape)
            img[rr, cc] += 512
            # plt.figure()
            self._dispnobl(img)

        self._params = model.params
        self._success = model.estimate(coords)

    def fit_circle_correlation(self,
                               method='Nelder-Mead',
                               display=False,
                               **keywords):
        '''Perform a circle fitting on the current mask using minimization 
        algorithm  with correlation merit functions.

        Tested with following minimizations methods: 'Nelder-Mead'. Relative 
        precision of 1% reached on synthetic images without noise.

        Parameters
        ----------
            method: string, default='Nelder-Mead'
                from scipy.optimize.minimize.
            display: bool, default=False
                SLOWLY shows the progress of the fit.
            **keywords: dict, optional
                passed to scipy.optimize.minimize
        '''

        self._method = 'correlation ' + method

        img = np.asarray(self._mask.copy(), dtype=int)
        regions = measure.regionprops(img)
        bubble = regions[0]

        x0, y0 = bubble.centroid
        r = bubble.major_axis_length / 2.
        if display:
            fign = plt.figure()

        self._shape_fitted = 'circle'
        self._initial_guess = (x0, y0, r)

        def _cost_disk(params):
            x0, y0, r = params
            coords = draw.disk((x0, y0), r, shape=img.shape)
            template = np.zeros_like(img)
            template[coords] = 1
            if display:
                self._dispnobl(template + img, fign)
            return -np.sum((template > 0) & (img > 0))

        res = optimize.minimize(_cost_disk, self._initial_guess,
                                method=method, **keywords)
        self._params = res.x
        self._success = res.success
        if res.success is False or (method != 'COBYLA' and res.nit == 0):
            raise Exception("Fit circle didn't converge %s" % res)

    def fit_annular_correlation(self,
                                method='Nelder-Mead',
                                display=False,
                                **keywords):
        '''Perform a annular circle fitting on the current mask using 
        minimization algorithm  with correlation merit functions.

        Tested with following minimizations methods: 'Nelder-Mead'. Relative 
        precision of 1% reached on synthetic images without noise.

        Parameters
        ----------
            method: string, default='Nelder-Mead'
                from scipy.optimize.minimize. Choose among 'Nelder-Mead'
            display: bool, default=False
                SLOWLY shows the progress of the fit.
            **keywords: dict, optional
                passed to scipy.optimize.minimize
        '''

        self._method = 'correlation ' + method
        img = np.asarray(self._mask.copy(), dtype=int)
        regions = measure.regionprops(img)
        bubble = regions[0]

        x0, y0 = bubble.centroid
        r = bubble.major_axis_length / 2.
        inr = r / 2
        if display:
            fign = plt.figure()

        self._shape_fitted = 'annulus'
        self._initial_guess = (x0, y0, r, inr)

        def _cost_annular_disk(params):
            x0, y0, r, inr = params
            coords = draw.disk((x0, y0), r, shape=img.shape)
            template = np.zeros_like(img)
            template[coords] = 1

            coords2 = draw.disk((x0, y0), inr, shape=img.shape)
            template2 = np.zeros_like(img)
            template2[coords2] = 1
            template -= template2

            if display:
                self._dispnobl(template + img, fign)

            merit_fcn = np.sum((template - img)**2)

            return np.sqrt(merit_fcn)

        linear_constraint = LinearConstraint(
            np.identity(4, float32), np.zeros(4),
            np.zeros(4) + np.max(img.shape))
        res = optimize.minimize(_cost_annular_disk, self._initial_guess,
                                method=method, constraints=linear_constraint,
                                **keywords)
        self._params = res.x
        self._success = res.success
        if res.success is False or (method != 'COBYLA' and res.nit == 0):
            raise Exception("Fit circle with hole didn't converge %s" % res)

    def _dispnobl(self, img, fign=None, **kwargs):

        if fign is not None:
            plt.figure(fign.number)
        plt.clf()
        plt.imshow(img, aspect='auto', **kwargs)
        plt.colorbar()
        plt.draw()
        plt.ion()
        plt.show()
        plt.pause(0.001)
        plt.ioff()




def factors(n):
    '''
    Returns a list of factors of *n*.
    No attempt at performance optimisation, use sparingly.
    '''
    factors=[]
    for i in range(2,n//2+1):
        if n % i==0:
           factors.append(i)
    return factors

def gcd(a,b):
    '''
    Returns the greatest common divisor of a and b.
    '''
    while b:
        a, b = b, a % b

    return a

def lcm(a,b):
    '''
    Returns the least common multiple of a and b.
    '''
    return (a*b) // gcd(a,b)




import code
import signal
import traceback

# From https://stackoverflow.com/questions/132058/showing-the-stack-trace-from-a-running-python-application
#
# Call listen() when starting up the application
# When the application hangs or you want to interrupt
# and debug it, send the SIGUSR1 signal:
#
# kill -SIGUSR1 <pid>

def debug(sig, frame):
    """Interrupt running process, and provide a python prompt for
    interactive debugging."""
    d={'_frame':frame}         # Allow access to frame object.
    d.update(frame.f_globals)  # Unless shadowed by global
    d.update(frame.f_locals)

    i = code.InteractiveConsole(d)
    message  = "Signal received : entering python shell.\nTraceback:\n"
    message += ''.join(traceback.format_stack(frame))
    i.interact(message)

def listen( s=signal.SIGUSR1):
    signal.signal(s, debug)  # Register handle

 




###############################################################################
#  who           when        what
#  --------      ---------   ------------
#  A. Riccardi   April 1995  Written
#  A. Riccardi   2014-03-01  extended to odd Sampling. Rewritten to simplify
#                            the code. Ratio can be any positive number,
#                            no longer constrained to be larger than Sampling.
#  A. Puglisi    2019-09-08  ported to Python, added dtype parameter
###############################################################################

import numpy as np


def make_xy(sampling, ratio, dtype=None, polar=False, vector=False,
            zero_sampled=False, quarter=False, fft=False):
    '''
    Generates zero-centered domains in a cartesian plane.

    Generates a zero-centered domain on a cartesian plane or axis,
    using cartesian or polar coordinates, tipically for pupil sampling
    and FFT usage.

    Calling sequence::

       x,y = make_xy(sampling, ratio)
       r, theta = make_xy(sampling, ratio, polar=True)

    Parameters
    ----------
    sampling: int
              Number of sampling points per dimension of the domain, greater
              or equal to 2.
    ratio: float
           Extension of sampled domain: -Ratio <= x [,y] <= +Ratio

    Other Parameters
    ----------------
    polar: bool, optional
           If True, return domain sampling in polar coordinates. Default
           value is False (use cartesian coordinates)
    dtype: np.dtype, optional
           If set, the result will have this dtype. Otherwise, it will be
           inferred by the dtype of *sampling* and *ratio*
    vector: bool, optional
            If True, 1-dimensional domain is sampled instead of 2d.
            If *polar* is True, the value of *vector* is ignored.
    zero_sampled: bool, optional
                  If True, origin of the domain is sampled. This
                  flag is useful to force the zero to be sampled
                  when sampling is even. When sampling is odd, the
                  zero is always sampled
    quarter: bool, optional
             If True, only 1st quadrant is returned with (X>=0 AND Y>=0).
             The array returned has:
             * if Sampling is even: Sampling/2 X Sampling/2 elements
             * if Sampling is odd:  (Sampling+1)/2 X (Sampling+1)/2 elements
    fft: bool, optional
         If True, order the output values for FFT purposes. For example:

         * Sampling=4, Ratio=1 vector -3/4,-1/4,+1/4,+3/4:
             returned as +1/4,+3/4,-3/4,-1/4.
         * Sampling=4, Ratio=1, zero_sampled=True, vector -1,-1/2,0,+1/2:
             returned as 0,1/2,-1,-1/2
         * Sampling=5, Ratio=1 vector -4/5,-2/5,0,+2/5,+4/5:
             returned as 0,+2/5,+4/5,-4/5,-2/5 

    Returns
    -------
      (X, Y): tuple of numpy arrays
            with no special options, a 2-elements tuple with the X and Y
            values of sampled points
      (R, Angle): tuple of numpy arrays
            if *polar* is True, a 2-elements tuple with radial and
            angular values (in radians) of sampled points
      X: numpy array
            if *vector* is True, numpy array with X values of sampled points.

    Raises
    ------
    ValueError
        If the sampling parameter is lower than 2.

    Notes
    -----
    HOW THE DOMAIN IS SAMPLED

    The concept is the following: considering an array of
    `sampling` x `sampling pixels, the bottom-left corner of the bottom-left
    pixel has coordinates (-`ratio`,-`ratio`) and the top-right cornet of the
    to-left pixel has coordinates (+`ratio`,+`ratio`). The procedure
    returns the coordinates of the centers of the pixels. When `sampling` is
    even and `zero_sampled` is True, the coordinates of the bottom-left corners
    of the pixels are returned.

    * `sampling` is even and `zero_sampled` is False: the edge of the domain
      is not sampled and the sampling is symmetrical respect to origin::

               Ex: Sampling=4, Ratio=1.
                   -1   -0.5    0    0.5    1    Domain (Ex. X axis)
                    |     |     |     |     |
                       *     *     *     *       Sampling points
                     -0.75 -0.25  0.25  0.75     Returned vector


    * `sampling` is even and `zero_sampled` is True: the lower edge is sampled
      and the sampling is not symmetrical respect to the origin::

              Ex: Sampling=4, Ratio=1.
                   -1   -0.5    0    0.5    1    Domain (Ex. X axis)
                    |     |     |     |     |
                    *     *     *     *          Sampling points
                   -1   -0.5    0    0.5         Returned vector

    * `sampling`  is odd (`zero_sampled`  is ignored): the zero is always
      sampled::

               Ex: Sampling=5, Ratio=1.
                   -1   -3/5  -1/5   1/5   3/5    1    Domain (Ex. X axis)
                    |     |     |     |     |     |
                       *     *     *     *     *       Sampling points
                     -4/5  -2/5    0    2/5   4/5      Returned vector

    * If `fft` is True, output values are ordered for FFT purposes::

           Ex: 2-dimensional domain: N = Sampling
           X or Y(0:N/2-1, 0:N/2-1)   1st quadrant (including origin
                                      if ZERO_SAMPLEDis set)
           X or Y(N/2:N, 0:N/2-1)     2nd quadrant
           X or Y(N/2:N, N/2:N)       3rd quadrant
           X or Y(0:N/2-1, N/2:N)     2nd quadrant

    Examples
    --------
    Compute a tilt plane on a round pupil

    >>> x, y = make_xy(256, 1.0)
    >>> pupil = x*0
    >>> pupil[(x*x + y*y)<=1] =1
    >>> plt.imshow(pupil)

    '''

    if sampling <= 1:
        raise ValueError('make_xy: sampling must be larger than 1')
    even_sampling = ((sampling % 2) == 0)

    if quarter:
        if even_sampling:
            size = sampling // 2
            if zero_sampled:
                x0 = 0.0
            else:
                x0 = -0.5
        else:
            size = (sampling + 1) // 2
            x0 = 0.0
    else:
        size = sampling
        x0 = (sampling - 1) / 2.0
        if even_sampling and zero_sampled:
            x0 += 0.5

    x = (np.arange(size, dtype=dtype) - x0) / (sampling / 2.0) * ratio

    if fft and not quarter:
        if even_sampling:
            x = np.roll(x, -sampling / 2)
        else:
            x = np.roll(x, -(sampling - 1) / 2)

    if vector:
        return x

    x = np.tile(x, (size, 1))
    y = np.transpose(x).copy()
    if polar:
        x, y = _xy_to_polar(x, y)

    return x, y


def _xy_to_polar(x, y):

    r = np.sqrt(x * x + y * y)
    y = np.arctan2(y, x)
    return r, y

# ___oOo___



from arte.utils.timestamp import Timestamp


class Tag(object):

    def __init__(self, tag_string):
        assert tag_string.count("_") > 0
        self._tag_string = tag_string

    def get_day_as_string(self):
        return self._tag_string.split("_")[0]

    def __str__(self):
        return self._tag_string

    @staticmethod
    def create_tag():
        '''Create a tag with the current timestamp'''
        return Tag(Timestamp().now())



'''
@author: giuliacarla
'''


import astropy.units as u
from astropy.units.quantity import Quantity


class CircularOpticalAperture():
    '''
    This class defines the geometry of a circular optical aperture.

    Parameters
    ----------
    aperture_radius: float
        Radius of the optical aperture in [m].

    cartes_coords: tuple of three floats
        Cartesian coordinates (x, y, z) of the aperture center
        in [m].
    '''

    def __init__(self, aperture_radius, cartes_coords):
        self._r = aperture_radius
        self._x = cartes_coords[0]
        self._y = cartes_coords[1]
        self._z = cartes_coords[2]

    def getCartesianCoords(self):
        if (isinstance(self._x, Quantity) and (self._y, Quantity)
                and (self._z, Quantity)):
            return [self._x.to(u.m), self._y.to(u.m), self._z.to(u.m)]
        else:
            return [self._x * u.m, self._y * u.m, self._z * u.m]

    def getApertureRadius(self):
        if (isinstance(self._r, Quantity)):
            return self._r.to(u.m)
        else:
            return self._r * u.m



import numpy as np
from arte.types.scalar_bidimensional_function import ScalarBidimensionalFunction
from arte.types.domainxy import DomainXY


class BidimensionalFourierTransform(object):

    @staticmethod
    def distances_x_map(sizeInPoints, pixelSize):
        domain = DomainXY.from_shape((sizeInPoints, sizeInPoints), pixelSize)
        return domain.xcoord

    @staticmethod
    def distances_norm_map(sizeInPoints, pixelSize):
        domain = DomainXY.from_shape((sizeInPoints, sizeInPoints), pixelSize)
        return np.linalg.norm(np.dstack((domain.xmap, domain.ymap)), axis=2)

    @staticmethod
    def frequencies_norm_map(sizeInPoints, pixelSize):
        n = sizeInPoints
        a = np.tile(np.fft.fftfreq(n, d=pixelSize), (n, 1))
        return np.fft.fftshift(np.linalg.norm(np.dstack((a, a.T)), axis=2))

    @staticmethod
    def frequencies_x_map(sizeInPoints, pixelSize):
        n = sizeInPoints
        a = np.tile(np.fft.fftfreq(n, d=pixelSize), (n, 1))
        return np.fft.fftshift(a)

    @staticmethod
    def frequencies_y_map(sizeInPoints, pixelSize):
        return BidimensionalFourierTransform.frequencies_x_map(
            sizeInPoints, pixelSize).T

    @staticmethod
    def _isEven(num):
        if num % 2 == 0:
            return True
        else:
            return False

    @staticmethod
    def most_positive_frequency(sizeInPoints, pixelSize):
        sizeInLengthUnits = sizeInPoints * pixelSize
        if BidimensionalFourierTransform._isEven(sizeInPoints):
            return (0.5 * sizeInPoints - 1) / sizeInLengthUnits
        else:
            return 0.5 * (sizeInPoints - 1) / sizeInLengthUnits

    @staticmethod
    def most_negative_frequency(sizeInPoints, pixelSize):
        sizeInLengthUnits = sizeInPoints * pixelSize
        if BidimensionalFourierTransform._isEven(sizeInPoints):
            return -0.5 * sizeInPoints / sizeInLengthUnits
        else:
            return -0.5 * (sizeInPoints - 1) / sizeInLengthUnits

    @staticmethod
    def smallest_frequency(sizeInPoints, pixelSize):
        sizeInLengthUnits = sizeInPoints * pixelSize
        return 1. / sizeInLengthUnits

    @staticmethod
    def direct_transform(data):
        res = np.fft.fftshift(
            np.fft.fft2(
                np.fft.fftshift(data, axes=(-1, -2)),
                norm="ortho"),
            axes=(-1, -2))
        return res

    @staticmethod
    def inverse_transform(data):
        res = np.fft.ifftshift(
            np.fft.ifft2(
                np.fft.ifftshift(data),
                norm="ortho"))
        return res

    @staticmethod
    def direct(xyFunct):
        sizeY, sizeX = xyFunct.values.shape
        pxSizeX, pxSizeY = xyFunct.domain.step
        return ScalarBidimensionalFunction(
            BidimensionalFourierTransform.direct_transform(
                xyFunct.values),
            xmap=BidimensionalFourierTransform.frequencies_x_map(
                sizeX, pxSizeX),
            ymap=BidimensionalFourierTransform.frequencies_y_map(
                sizeY, pxSizeY)
        )

    @staticmethod
    def reverse(xyFunct):
        sizeY, sizeX = xyFunct.values.shape
        pxSizeX, pxSizeY = xyFunct.domain.step
        return ScalarBidimensionalFunction(
            BidimensionalFourierTransform.inverse_transform(
                xyFunct.values),
            xmap=BidimensionalFourierTransform.frequencies_x_map(
                sizeX, pxSizeX),
            ymap=BidimensionalFourierTransform.frequencies_y_map(
                sizeY, pxSizeY)
        )

#     @staticmethod
#     def directRealTransform(data, spacing):
#         res= np.fft.fftshift(
#             np.fft.rfft2(
#                 np.fft.fftshift(data, axes=(-1, -2))),
#             axes=(-1, -2)) * spacing**2
#         return res
#
#
#     @staticmethod
#     def inverseRealTransform(data, frequencySpacing):
#         sz= data.shape[0]
#         res= np.fft.ifftshift(
#             np.fft.irfft2(
#                 np.fft.ifftshift(data))) * (sz * frequencySpacing)**2
#         return res




'''
Created on Oct 13, 2019

@author: lbusoni
'''
import matplotlib.pyplot as plt
import synphot
from astropy import units as u
from synphot import SourceSpectrum
from synphot.models import Box1D, Empirical1D
from synphot.spectrum import SpectralElement
from synphot.observation import Observation
from arte.photometry.filters import Filters
from synphot.units import FLAM
from arte.photometry import get_normalized_star_spectrum
from arte.photometry.eso_sky_calc import EsoSkyCalc
import numpy as np
from arte.photometry.transmissive_elements import TransmissiveElement, Bandpass


def misc():
    # get_vega() downloads this one
    synphot.specio.read_remote_spec(
        'http://ssb.stsci.edu/cdbs/calspec/alpha_lyr_stis_008.fits')

    # G5V of UVKLIB subset of Pickles library
    # see http://www.stsci.edu/hst/instrumentation/reference-data-for-calibration-and-tools/astronomical-catalogs/pickles-atlas.html
    synphot.specio.read_remote_spec(
        'http://ssb.stsci.edu/cdbs/grid/pickles/dat_uvk/pickles_uk_27.fits')

    # read the sourcespectrum
    spG5V = SourceSpectrum.from_file(
        'http://ssb.stsci.edu/cdbs/grid/pickles/dat_uvk/pickles_uk_27.fits')
    spG2V = SourceSpectrum.from_file(
        'http://ssb.stsci.edu/cdbs/grid/pickles/dat_uvk/pickles_uk_26.fits')

    filtR = SpectralElement.from_filter('johnson_r')
    spG2V_19r = spG2V.normalize(
        19 * synphot.units.VEGAMAG, filtR,
        vegaspec=SourceSpectrum.from_vega())

    bp = SpectralElement(Box1D, x_0=700 * u.nm, width=600 * u.nm)
    obs = Observation(spG2V_19r, bp)
    obs.countrate(area=50 * u.m ** 2)

    bp220 = SpectralElement(Box1D, x_0=800 * u.nm, width=400 * u.nm)
    bpCRed = SpectralElement(Box1D, x_0=1650 * u.nm, width=300 * u.nm) * 0.8

    Observation(spG2V_19r, bp220).countrate(area=50 * u.m ** 2)
    Observation(spG2V_19r, bpCRed).countrate(area=50 * u.m ** 2)

    spM0V_8R = get_normalized_star_spectrum('M0V', 8.0, 'johnson_r')

    uPhotonSecM2Micron = u.photon / (u.s * u.m ** 2 * u.micron)

    spG2V_8R = get_normalized_star_spectrum("G2V", 8.0, 'johnson_r')
    plt.plot(spG2V_8R.waveset, spG2V_8R(spG2V_8R.waveset).to(uPhotonSecM2Micron))

    # compare with Armando's
    spG2V_19R = get_normalized_star_spectrum("G2V", 19, 'johnson_r')
    bp = SpectralElement(Box1D, x_0=700 * u.nm, width=600 * u.nm)
    obs = Observation(spG2V_19R, bp)
    obs.countrate(area=50 * u.m ** 2)

    # zeropoint in filtro r in erg/s/cm2/A
    Observation(get_normalized_star_spectrum('A0V', 0, 'johnson_r'), SpectralElement.from_filter('johnson_r')).effstim('flam')
    # zeropoint in ph/s/m2
    Observation(get_normalized_star_spectrum('A0V', 0, 'johnson_r'), SpectralElement.from_filter('johnson_r')).countrate(area=1 * u.m ** 2)


def check_zeropoints_ESO():
    '''
    Shouldn't they match http://www.eso.org/observing/etc/doc/skycalc/helpskycalc.html#mags?
    '''
    # sky = EsoSkyCalc(airmass=1, moon_target_sep=45.0, moon_alt=45,
    #                 observatory='paranal', wdelta=10)
    # atmo = SpectralElement(Empirical1D, points=sky.lam, lookup_table=sky.trans)

    filters = [Filters.ESO_ETC_U,
               Filters.ESO_ETC_B,
               Filters.ESO_ETC_V,
               Filters.ESO_ETC_R,
               Filters.ESO_ETC_I,
               Filters.ESO_ETC_Z,
               Filters.ESO_ETC_Y,
               Filters.ESO_ETC_J,
               Filters.ESO_ETC_H,
               Filters.ESO_ETC_K,
               Filters.ESO_ETC_L,
               Filters.ESO_ETC_M
               ]

    eso_etc_zp = np.array([
        4.18023e-09, 6.60085e-09, 3.60994e-09,
        2.28665e-09, 1.22603e-09, 7.76068e-10,
        5.973e-10, 3.12e-10, 1.14e-10, 3.94e-11,
        4.83e-12, 2.04e-12, 1.23e-13, 6.8e-15]) * FLAM

    star = get_normalized_star_spectrum('vega', 0.0, Filters.ESO_ETC_R)
    print(f"{'filter':10s} | {'flux':15s} | {'zp err %':8s} | {'cnts':17s} |"
          f"{'boundaries'}")
    for filt_name, etc_zp in zip(filters, eso_etc_zp):
        filt = Filters.get(filt_name)
        obs = Observation(star, filt)
        zp = obs.effstim('flam', wavelengths=filt.waveset)
        err = (zp - etc_zp) / etc_zp * 100
        cnts = obs.countrate(area=1 * u.m ** 2, wavelengths=filt.waveset)
        print(f"{filt_name:10s} | {zp:10.3e} | {err:+7.3f}% | {cnts:10.3e} |"
              f"{filt.waveset.to(u.nm)[0]:g} {filt.waveset.to(u.nm)[-1]:g}")



def pippo(filt):
    f_source = get_normalized_star_spectrum(spectral_type='vega', magnitude=0, filter_name=filt)
    obs = Observation(spec=f_source, band=Filters.get(filt), binset=f_source.waveset)
    return obs.countrate(area=1*u.m**2)


def compare_integrate_with_observation(filt_name, filt_minmax=None, filt_obs=None, star_spectral_type='A0V'):
    sp = get_normalized_star_spectrum(star_spectral_type, 0, filt_name)
    filt = Filters.get(filt_name)
    if filt_minmax is None:
        waveset = filt.waveset
    else:
        waveset = np.linspace(filt_minmax[0], filt_minmax[1], 1000)
    counts_int = sp.integrate(waveset) * 1e4 * u.cm**2/u.m**2
    if filt_obs == 'one':
        #Observation integrates the spectrum considering 100% transmission within the band
        filt_obs = SpectralElement(
            Box1D, amplitude=1., x_0=(waveset.max() + waveset.min())/2, width=(waveset.max() - waveset.min()))
    elif filt_obs is None:
        filt_obs = filt
    obs = Observation(spec=sp, band=filt_obs, binset=waveset, force='taper')
    counts_obs = obs.countrate(area=1 * u.m**2)
    print('Flux with integrate function: %s' %counts_int)
    print('Flux with Observation: %s' %counts_obs)
    return sp


def main_observation_playing_with_parameters():
    print('Integrate considering a waveset built by us with minmax of our H band and a Box1D filter:')
    sp = compare_integrate_with_observation(filt_name=Filters.ESO_ETC_H, filt_minmax=(14900, 17800), filt_obs='one', star_spectral_type='vega')

    print('\nIntegrate considering the ESO filter waveset (min is smaller and max is larger) and a Box1D filter:')
    sp = compare_integrate_with_observation(filt_name=Filters.ESO_ETC_H, filt_obs='one', star_spectral_type='vega')



import synphot
from synphot import SourceSpectrum
from arte.photometry.spectral_types import PickelsLibrary
from arte.photometry.filters import Filters


def get_normalized_star_spectrum(spectral_type, magnitude, filter_name):
    """
    spec_data = get_normalized_star_spectrum(spectral_type, magnitude, filter_name)

    Returns a structure containing the synthetic spectrum of the star having the spectral type and magnitude 
    in the specified input filter. Magnitude is in VEGAMAG-F(lambda) system.
    Spectra are from PICKLES, PASP, 110, 863 (1998)
    Absolute flux spectra, no effect of atmospheric and instrument transmission


    Parameters
    ----------
        r0AtZenith: float
                    overall r0 at zenith [m]

        spectral_type:  string.
                        spectral type and luminosity class (e.g. G2V or M4III) or 'vega'
        magnitude:  float.
                    magnitude in the filter_name filter
        filter_name: string.
                     Name of the filter. See Filters.get() for the list of available filters

    Returns
    -------
        spectrum: synphot.SourceSpectrum object defining the spectrum

    Examples
    --------
    Plot the spectrum of a vega, A0V, G2V stars of mag=8 defined on JohnsonR filter

    >>> sp= get_normalized_star_spectrum('vega', 8, Filters.JOHNSON_R)
    >>> spA0V= get_normalized_star_spectrum('A0V', 8, Filters.JOHNSON_R)
    >>> spG2V= get_normalized_star_spectrum('G2V', 8, Filters.JOHNSON_R)
    >>> plt.plot(sp.waveset, sp(sp.waveset), label='Vega')
    >>> plt.plot(spA0V.waveset, spA0V(spA0V.waveset), label='A0V')
    >>> plt.plot(spG2V.waveset, spG2V(spG2V.waveset), label='G2V')
    >>> plt.grid(True)
    >>> plt.xlabel('nm')
    >>> plt.ylabel('FLAM')
    >>> plt.xlim(0, 10000)
    >>> plt.legend()
"""
    # read the sourcespectrum
    if spectral_type == 'vega':
        spectrum = SourceSpectrum.from_vega()
    else:
        spectrum = SourceSpectrum.from_file(
            PickelsLibrary.filename(spectral_type))

    bandpass = Filters.get(filter_name)

    spectrum_norm = spectrum.normalize(
        magnitude * synphot.units.VEGAMAG,
        bandpass,
        vegaspec=SourceSpectrum.from_vega())

    return spectrum_norm



##########################################################
#
# who       when        what
# --------  ----------  ----------------------------------
# apuglisi  2020-04-23  Reimplemented as a class decorator
# apuglisi  2019-09-28  Created
#
##########################################################
'''
Provides tools to build an interactive, searchable help based on docstrings.

Any class to decorated with @add_help
gets a :meth:`help` method that provides an interactive and searchable help
based on the method docstrings. All public methods (not starting with "_")
are added, together with all such methods in all the members that are
classes decorated with @add_help. Derived classes inherit the help system
without a need to use the @add_help decorator.

Help is built dynamically when invoked, so if a new member is added
to the class at runtime, it will appear in the help too.

The help method name can be customized giving the `help_function` parameter
to the decorator.

If the `classmethod` parameter is True, the help function is created
as a classmethod instead of an ordinary method. This is useful
for classes that only define classmethods and are not normally instanced.

Example::

  from arte.utils.help import add_help

  @add_help
  class InnerClass():
      """An inner class"""

      def a_method(self):
          """This is a method"""
          pass

  @add_help
  class MyClass():
      """This is my class"""
      b = InnerClass()

      def a_method(self):
          """This is a method"""
          pass

      def another_method(self):
          """This is another method"""
          pass

  @add_help(help_function='show_help')
  class CustomClass():
      """This a custom class"""

      def a_method(self):
          """This is a method"""
          pass

Interactive session::

  >>> a  = MyClass()
  >>> a.help()
  ---------
  MyClass                    This is my class
  MyClass.a_method()         This is a method
  MyClass.another_method()   This is another method
  ---------
  MyClass.b              An inner class
  MyClass.b.a_method()   This is a method

  >>> a.help('other')
  ---------
  MyClass.another_method()   This is another method

  >>> a.help('inner')
  ---------
  MyClass.b              An inner class

  >>> b = CustomClass()
  >>> b.show_help()
  ---------
  CustomClass               This a custom class
  CustomClass.a_method()    This is a method

'''
import inspect
from functools import partial
from functools import cached_property

HIDDEN_HELP = '__arte_help'
HIDDEN_NAME = '__arte_name'
HIDDEN_ARGS = '__arte_args'
HIDDEN_HIDE = '__arte_hide'


def _is_public_method(name):
    return name[0] != '_'


def _is_hlp_class(obj):
    return hasattr(obj, HIDDEN_HELP)


def _is_hidden(m):
    return (hasattr(m, HIDDEN_HIDE)) and (getattr(m, HIDDEN_HIDE) is True)


def add_help(cls=None, *, help_function='help', classmethod=False):
    '''
    Decorator to add interactive help to a class

    Parameters
    ----------
    help_function: str, optional
        Name of the method that will be added to the class. Defaults to "help"
    classmethod: bool, optional
        If True, the help method will be added as a classmethod. Default False

    Returns
    -------
    class
        The decorated class type
    '''
    # Trick to allow a decorator without parenthesis
    if cls is not None:
        return add_help()(cls)

    def help(self, search='', prefix=''):
        '''
        Interactive help

        Prints on stdout a list of methods that match the *search* substring
        or all of them if *search* is left to the default value of an empty
        string, together with a one-line help taken from the first line
        of their docstring, if any.

        The *prefix* argument is prepended to the method name and is used
        for recursive help of every class member.
        '''
        properties = ({k: getattr(self.__class__, k)
                      for k in dir(self.__class__)
                      if isinstance(getattr(self.__class__, k), (property, cached_property))})

        items = [k for k in dir(self) if k not in properties]

        methods = {k: getattr(self, k) for k in items if callable(getattr(self, k))}
        members = {k: getattr(self, k) for k in items if not callable(getattr(self, k))}
        methods.update(properties)

        methods = {k: v for k, v in methods.items() if _is_public_method(k) \
                                                      and not _is_hidden(v)}
        members = {k: v for k, v in members.items() if _is_hlp_class(v)}

        if prefix == '':
            prefix = self.__class__.__name__

        hlp = {prefix: _format_docstring(self)}

        for name, method in methods.items():
            if name and method:  # Skip NAs
                display_name = _format_name(method, default=name)
                if name in properties:
                    pars = ''
                else:
                    pars = _format_pars(method)
                helpstr = _format_docstring(method)
                hlp[prefix + '.' + display_name + pars] = helpstr

        maxlen = max(map(len, hlp.keys()))
        fmt = '%%-%ds%%s' % (maxlen + 3)

        lines = []
        for k in sorted(hlp.keys()):
            line = fmt % (k, hlp[k])
            if search in line:
                lines.append(line)

        if len(lines) > 0:
            print('---------')
            for line in lines:
                print(line)

        for name, obj in sorted(members.items()):
            if name[0] != '_':
                obj.help(search=search, prefix='%s.%s' % (prefix, name))

    def decorate(cls):
        setattr(cls, HIDDEN_HELP, False)  # Set attr but do not define a string
        if classmethod:
            func = partial(help, self=cls())
            func.__name__ = help.__name__
            setattr(cls, help_function, hide_from_help(func))
        else:
            setattr(cls, help_function, hide_from_help(help))
        return cls
    return decorate


def modify_help(call=None, arg_str=None, doc_str=None):
    '''
    Decorator to modify the automatic help for a method.

    Without this decorator, the method signature for help
    is just "method()". Using this decorator, other
    signatures are possible::

      @modify_help(call='mymethod1(foo)')
      def mymethod1(self, ....)
          """This method is very cool"""

      @modify_help(arg_str='idx1, idx2')
      def mymethod2(self, ....)
          """Now you see it"""

      @modify_help(doc_str='Surprise!')
      def mymethod3(self, ....)
          """And now you don't"""

    Resulting help::

      .mymethod1(foo)        : This method is very cool
      .mymethod2(idx1, idx2) : Now you see it
      .mymethod3()           : Surprise!

    .. Note::
        if the method is a @staticmethod, this decorator
        should be inserted *after* the staticmethod one.
    '''
    def wrap(f):
        _wrap_with(f, call, arg_str, doc_str)
        return f
    return wrap


def hide_from_help(f):
    '''Decorator to hide a method from the interactive help'''
    setattr(f, HIDDEN_HIDE, True)
    return f


def _format_docstring(obj, default=None):

    hlp = getattr(obj, HIDDEN_HELP, False)
    docstr = (obj.__doc__ or '').strip()
    hlp = hlp or docstr or default or 'No docstring defined'
    hlp = hlp.strip().splitlines()[0]
    return hlp


def _format_name(obj, default=None):

    if hasattr(obj, '__name__'):
        myname = obj.__name__
    else:
        myname = default

    name = getattr(obj, HIDDEN_NAME, False)
    name = name or myname
    name = name.strip().splitlines()[0]
    return name


def _format_pars(method):
    if isinstance(method, property):
        return ''

    args = getattr(method, HIDDEN_ARGS, None)
    if args is not None:
        return args

    sig = inspect.signature(method)
    return '(' + ','.join(sig.parameters.keys()) + ')'


def _wrap_with(f, call=None, arg_str=None, doc_str=None):

    if call:
        name = call
        args = ''
    else:
        name = f.__name__
        if arg_str:
            args = '(%s)' % arg_str
        else:
            args = None

    hlp = doc_str or _format_docstring(f)
    setattr(f, HIDDEN_NAME, name)
    setattr(f, HIDDEN_HELP, hlp)
    setattr(f, HIDDEN_ARGS, args)

# ___oOo___
